package co.id.datacomsolusindo.ipphonebridge

import com.fasterxml.jackson.core.type.TypeReference
import com.fasterxml.jackson.databind.ObjectMapper
import org.apache.commons.io.IOUtils
import org.springframework.core.Ordered
import org.springframework.core.annotation.Order
import org.springframework.http.HttpMethod
import org.springframework.http.HttpStatus
import org.springframework.messaging.simp.SimpMessagingTemplate
import org.springframework.stereotype.Component
import org.springframework.web.bind.annotation.GetMapping
import org.springframework.web.bind.annotation.PathVariable
import org.springframework.web.bind.annotation.PostMapping
import org.springframework.web.bind.annotation.RestController
import java.io.IOException
import java.io.InputStream
import java.io.Serializable
import java.net.InetAddress
import java.util.*
import java.util.concurrent.TimeUnit
import javax.servlet.*
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import kotlin.streams.toList


@Component
@Order(Ordered.HIGHEST_PRECEDENCE)
class SimpleCORSFilter : Filter {
    @Throws(ServletException::class)
    override fun init(fc: FilterConfig?) {
    }

    @Throws(IOException::class, ServletException::class)
    override fun doFilter(req: ServletRequest, resp: ServletResponse,
                          chain: FilterChain) {
        val response = resp as HttpServletResponse
        val request = req as HttpServletRequest
        response.setHeader("Access-Control-Allow-Origin", "*")
        response.setHeader("Access-Control-Allow-Methods", "PATCH,POST,GET,OPTIONS,DELETE")
        response.setHeader("Access-Control-Max-Age", "3600")
        response.setHeader("Access-Control-Allow-Headers", "x-requested-with, authorization, Content-Type, Authorization, credential, X-XSRF-TOKEN")
        if ("OPTIONS".equals(request.method, ignoreCase = true)) {
            response.status = HttpServletResponse.SC_OK
        } else {
            chain.doFilter(req, resp)
        }
    }

    override fun destroy() {}
}

@Component
@Order(1)
class BridgeFilter(private val template: SimpMessagingTemplate) : Filter {
    override fun doFilter(request: ServletRequest, response: ServletResponse, chain: FilterChain) {
        val req = request as HttpServletRequest
        val splitPath = req.requestURI.split("/")
        val client = splitPath[1]
        val toPath = "/" + splitPath.takeLast(splitPath.size - 1).joinToString("/")
        if (client.startsWith("_")) {
            chain.doFilter(request, response)
            return
        }

        println("client $client")
        println("path $toPath")

//        req.parameterMap.entries.forEach {
//            println(it.key + " : " + it.value.joinToString(", "))
//        }

        val headerMap = req.headerNames.toList().associateBy({ it }, { req.getHeader(it) })
        val body = req.reader.lines().toList().joinToString(System.lineSeparator())

        val reqId = UUID.randomUUID().toString()
        val rb = RequestBuilder(reqId, toPath, HttpMethod.valueOf(req.method), headerMap.toMutableMap())
        rb.body = body

        rb.queryString = req.parameterMap
        Singletons.requestInstance[reqId] = RequestQue(reqId, rb, null)

        template.convertAndSend("/topic/$client", RequestTrigger(reqId))

        val res = response as HttpServletResponse
        var i = 0

        while (Singletons.responseQue[reqId] == null) {
            TimeUnit.MILLISECONDS.sleep(100)
            i++
            if (i >= 600) {

                res.status = HttpStatus.REQUEST_TIMEOUT.value()
                res.writer.write("{\"message\":\"Request timeout. Client not responding\"}")
                res.flushBuffer()
                return
            }
        }

        val resFromSocket = Singletons.responseQue[reqId]
        res.status = resFromSocket!!.statusCode
        resFromSocket.headers?.let {
            it.entries.forEach { en ->
//                println(en.key + " : " + en.value.joinToString(","))
                res.setHeader(en.key, en.value.joinToString(","))
            }
        }
        res.outputStream.write(IOUtils.toByteArray(resFromSocket.body))
//        resFromSocket.body.let { res.writer.write(it) }
        res.flushBuffer()
        Singletons.requestInstance.remove(reqId)
        return
    }
}

@Component
@Order(Ordered.LOWEST_PRECEDENCE)
class RequestFilter : Filter {
    override fun doFilter(sreq: ServletRequest?, sres: ServletResponse?, fc: FilterChain?) {
        val req = sreq as HttpServletRequest
        println("Request With " + req.getHeader("User-Agent") + " Address " + InetAddress.getLocalHost().hostAddress + " To " + req.requestURL.toString() + "?" + req.queryString)
        fc!!.doFilter(sreq, sres)
    }
}

class RequestQue(val id: String, val requestBuilder: RequestBuilder, var responseObj: Resp?)

class RequestTrigger(val id: String)

class RequestBuilder(val id: String, val path: String, val method: HttpMethod, var headers: MutableMap<String, String>) : Serializable {
    var body: String? = null
    var queryString: MutableMap<String, Array<String>>? = null
}

class Resp(val body: InputStream, val statusCode: Int, val headers: Map<String,Array<String>>? = null)

object Singletons {
    val responseQue: MutableMap<String, Resp> by lazy { mutableMapOf<String, Resp>() }
    val requestInstance: MutableMap<String, RequestQue> by lazy { mutableMapOf<String, RequestQue>() }
}

@RestController
class BridgeRestController {
    @PostMapping("/_response/{id}")
    fun responseFromRest(@PathVariable("id") id: String, req: HttpServletRequest): String {
        val objectMapper = ObjectMapper()
        val typeRef: TypeReference<HashMap<String, Array<String>>> = object : TypeReference<HashMap<String, Array<String>>>() {}
        Singletons.responseQue[id] = Resp(
                req.getPart("body").inputStream,
               String(IOUtils.toByteArray(req.getPart("status").inputStream)).toInt(),
                if (req.getPart("header") != null) {
                    objectMapper.readValue(String(IOUtils.toByteArray(req.getPart("header").inputStream)), typeRef)
                } else {
                    null
                }
        )

//
//        IOUtils.toByteArray(req.getPart("body").inputStream)
//        println(req.getPart("header"))
//        println(req.getPart("status"))

        return "{\"success\":true}"
    }

    @GetMapping("/_request/{id}")
    fun getResponseObj(@PathVariable("id") id: String): RequestBuilder {
        if (Singletons.requestInstance[id] == null) {
            throw Exception("not found")
        }
        return Singletons.requestInstance[id]!!.requestBuilder
    }
}

